# Profiling IPEX on Sunspot/Aurora

Documenting ways to profile [IPEX](https://github.com/intel/intel-extension-for-pytorch) (Intel Extension for PyTorch) on Aurora/Sunspot using an adaptation of [an IPEX example](https://intel.github.io/intel-extension-for-pytorch/xpu/latest/tutorials/examples.html#training) from IPEX's documentation (saved as ipex_example.py). 

## Basic Usage
from Sunspot:
```
git clone https://git.cels.anl.gov/blusch/ipex_profiling.git
cd ipex_profiling
qsub submit_ipex_example.sh
```
This runs `ipex_example.py` on one tile without profiling. The first time you run it, it downloads the data. The code takes about 2 minutes even if you are downloading the data. (This calls `run_ipex_example.sh` as an intermediate step for the sake of using VTune later.)

If you're on a node without outbound network connectivity, you can [add proxy settings](https://alcf.anl.gov/support-center/aurora/getting-started-sunspot#SunspotGettingStartedGuide-ProxySettings) to your ~/.bash_profile file. 


## line_profiler
Line-by-line profiling of Python functions

You can see basic usage in the documentation: https://github.com/pyutils/line_profiler 

For example:
- You could replace `python ipex_example.py` with `python -m kernprof -l ipex_example.py` in a runner script. 
- Then you could add `from line_profiler import profile` to `ipex_example.py` and decorate a training step function with `@profile`.  
- Then, if the results are saved as `ipex_example.py.lprof`, you could look at the results with `python -m line_profiler ipex_example.py.lprof`. (You can do this on a login node if you load the frameworks module first.)

For convenience, you can try these edits by commenting and uncommenting lines in `ipex_example.py` and `run_ipex_example.sh`. This still runs for only about 2 minutes. 

We can immediately see room for massive improvement: 50% of the training loop time is spent moving the data to the GPU, and 11% is spent printing the status. 


Set up on Sunspot:\
This module is not included in the Sunspot conda environment from `module load frameworks/2023.12.15.001`. However, you could install it locally, for example with `pip install --user line_profiler`, or use the venv approach recommended [here](https://docs.alcf.anl.gov/aurora/data-science/python/).

## cProfile
cProfile is part of the Python standard library and times function calls. 

Documentation: https://docs.python.org/3/library/profile.html

For example:
- You could replace `python ipex_example.py` with `python -m cProfile -o profile_1gpu ipex_example.py` in a runner script to profile the entire Python script. (The part after -o is the name you want for the output file.) To investigate the output file in Python later, you can do things like this: 
```
import pstats
p = pstats.Stats(filename)
p.sort_stats('cumtime').print_stats(30)
```
- To profile just one training step instead of everything, you want to *instead* edit `ipex_example.py`, for example like this:

`import cProfile`

and then inside the loop over batches:
```
    if batch_idx == 42:
        with cProfile.Profile() as pr:
            train_step(data, target, batch_idx)
        pr.print_stats('cumulative')
        pr.print_stats('tottime')
    else:
        train_step(data, target, batch_idx)
```
- The above changes to `ipex_example.py` included printing some stats immediately, but you could also save the profile using `pr.dump_stats(filename).`


For convenience, you can try these edits by commenting and uncommenting lines in `ipex_example.py` and `run_ipex_example.sh`. Both approaches only run for about two minutes. 

## PyTorch Profiler 
Measures time and memory consumption of model operators. (Note that this is only for PyTorch functions.)\
Tutorial: https://pytorch.org/tutorials/recipes/recipes/profiler_recipe.html

### PyTorch Legacy Profiler

Important caveat with the legacy profiler: The printed table includes columns about XPU time. However, the Chrome trace does not contain information about what happens on the XPU. 

One option is to use PyTorch's legacy profiler API, which is under `torch.autograd`. See https://pytorch.org/docs/stable/autograd.html#profiler. 
For example:
- You could add this around `train_model()` in `ipex_example.py`:
```
with torch.autograd.profiler_legacy.profile(
    use_xpu=True, profile_memory=True, record_shapes=True, with_stack=True,
    ) as prof:
    ...
```
- with_stack=True includes the file and line number of operations in the trace, so it adds overhead but could be helpful.  
- record_shapes=True might skew the timings, but it collects information about input dimensions 
- You could then print a sorted table of timings. (This version of `ipex_example.py` ran for about 6 minutes.) If you later want to explore other ways to sort the table, you could copy it into something like Excel. 
```
print(prof.key_averages().table(sort_by="self_cpu_time_total"))
```
- You could also choose to save a Chrome trace. You can view the Chrome trace in the Chrome browser by going to chrome://tracing and loading the trace file. 
```
prof.export_chrome_trace("trace_file.json")
```
- In the case of this example, the Chrome trace is 360 MB, which is fine. However, the trace files saved this way are often too large to load. You'll either get an error or it will quietly fail without a trace appearing. (Perhaps one solution would be splitting the JSON file into smaller segments before loading it.) 
- You could also add the PyTorch profiler elsewhere in the code in order to save smaller traces. For example, the profiler could be around just one training step (`train_step(data, target, batch_idx)`). If, for your application, the trace is still too big to load in Chrome, you might want to wrap one forward pass (`output = model(data)`) and one backward pass (`loss.backward()`) separately so that two different trace files get saved.  

For convenience, you can try these edits by commenting and uncommenting lines in ipex_example.py. Whether I printed the table or saved the Chrome trace, the whole job took about six minutes.

Notes:
- If I used the legacy PyTorch profiler on Sunspot to create a Chrome trace and loaded it in Chrome, I got a JSON parsing error. I fixed it using this: 
`sed 's/{{}}}}/{}}/g' trace_file.json > trace_file_repaired.json`
- You may want to use `torch.profiler.record_function` so that regions of the code will be conveniently annotated. 

### PyTorch Profiler (newer)

The newer PyTorch profiler is documented [here](https://pytorch.org/docs/2.3/profiler.html#module-torch.profiler), and there is a recipe [here](https://pytorch.org/tutorials/recipes/recipes/profiler_recipe.html). You can use almost the same syntax as we used with the legacy profiler above, but using `torch.profiler.profile` as the context manager. 

The [Kineto](https://github.com/pytorch/kineto) project is part of the PyTorch profiler and enables tracking GPUs. 
The `torch.profiler.profile` function doesn't use parameters like `use_xpu`. Instead, the `activities` parameter can support a list of activities to track. We are interested in these two options: `torch.profiler.ProfilerActivity.CPU` and `torch.profiler.ProfilerActivity.XPU`. 

**Edit 1:** `IPEX_ZE_TRACING=1`\
 Intel has instructions for using the prototype "Kineto Supported Profiler Tool" [here](https://intel.github.io/intel-extension-for-pytorch/xpu/latest/tutorials/features/profiler_kineto.html). The key detail is we need to set the environment variable `IPEX_ZE_TRACING=1` before running PyTorch to enable the level zero tracing layer.

**Edit 2:** schedule function and prof.step()\
If we trace all of `train_model()` in `ipex_example.py` with this profiler, the trace is far too big to load in Chrome. However, we can specify a schedule for the profiler. (Read more in [the profiler recipe](https://pytorch.org/tutorials/recipes/recipes/profiler_recipe.html).) For example, this schedule will skip the first 10 steps, then start 1 cycle (repeat=1). The cycle consists of waiting for 5 steps, warming up for 1 step, and actively tracing for 3 steps. 
```
my_schedule = torch.profiler.schedule(skip_first=10, wait=5, warmup=1, active=3, repeat=1)
```
We need to add `prof.step()` in our training loop so the profiler can track the steps.

**Edit 3:** trace_handler\
Following the profiler recipe, we can also conveniently write a function that handles printing the table and saving a trace after a cycle ends.
```
def trace_handler(p):
    output = p.key_averages().table(sort_by="self_cpu_time_total", row_limit=20)
    print(output)
    p.export_chrome_trace("trace_step" + str(p.step_num) + ".json")
```
**Edit 4:** profile context manager with torch.profiler.profile...\
Putting this all together:
```
with torch.profiler.profile(
    activities=[torch.profiler.ProfilerActivity.CPU, torch.profiler.ProfilerActivity.XPU],
    schedule=my_schedule,
    on_trace_ready=trace_handler
    ) as prof:
    train_step(data, target, batch_idx)
    prof.step()
```
For convenience, you can try these edits by commenting and uncommenting lines in `ipex_example.py` and `run_ipex_example.sh`. It runs in a few minutes.

## VTune
[VTune](https://www.intel.com/content/www/us/en/developer/tools/oneapi/vtune-profiler.html#gs.9ocuuy) is a profiler by Intel. There is a tutorial specifically on using VTune to profile PyTorch [here](https://pytorch.org/tutorials/recipes/profile_with_itt.html). VTune provides a Instrumentation and Tracing Technology (ITT) API, which is integrated into PyTorch. We can use it to label parts of the code. Following the tutorial, we can add ITT labeling this way: 
```
max_batches = 5
with torch.autograd.profiler.emit_itt():
    for batch_idx, (data, target) in enumerate(train_loader):
        with torch.profiler.itt.range(f'iteration_{batch_idx}'):
            train_step(data, target, batch_idx)
        if batch_idx >= max_batches:
            break
```
You might want to run for a limited number of steps, since VTune logs are extensive. For this example, if we only profile 5 batches, the finalization will finish within a 10-minute job. 

The PyTorch tutorial explains launching the application with the VTune GUI, but for Sunspot, it would likely be easier to use VTune programmatically. The tutorial mentions that "To profile a PyTorch script, it is recommended to wrap all manual steps, including activating a Python environment and setting required environment variables, into a bash script, then profile this bash script." We can handle this by having most steps in `run_ipex_example.sh.` The script `submit_ipex_example.sh` can handle PBS and launching `run_ipex_example.sh`. Here we need a sepcial environment variable and then can use VTune on `run_ipex_example.sh`: 
```
export AMPLXE_EXPERIMENTAL=gpu-multi-tile-metrics
ZE_AFFINITY_MASK=0.0 vtune -collect gpu-hotspots -result-dir=./vtune_log/ -- bash run_ipex_example.sh
```

After the job ends, we want to view the results in VTune. See the section "After collecting the performance data, VTune profiler web server can be used for the post-processing" in the [Aurora documentation for VTune](https://docs.alcf.anl.gov/aurora/performance-tools/vtune/#after-collecting-the-performance-data-vtune-profiler-web-server-can-be-used-for-the-post-processing). 

For convenience, you can try these edits by commenting and uncommenting lines in `submit_ipex_example.sh` and `ipex_example.py`.
